import numpy as np
import tensorflow as tf
from tensorflow.keras import  layers
import argparse
import os
import time


# using P1 network with adaptive mesh size
# Grid range is [-1,1]^d 
class P1AdaptLayer(layers.Layer):
    # dimIn  integer  : dimension of input
    # dimOut integer  : dimension of output
    # nbMesh int      : number of mesh per dimension
    def __init__(self,
                 in_features,
                 out_features,
                 nbMesh):
        super().__init__()
        self.in_features=in_features
        self.out_features= out_features
        self.nbMesh = nbMesh
        # unknown  are P1 function value
        self.weightsNet= self.add_weight(shape= (out_features,in_features,nbMesh+1), initializer="random_normal",trainable=True , dtype=tf.float32)
        # trainables are the mesh point on [0,1] inside the grid
        self.weiGhtaPt=  self.add_weight(shape= (in_features,nbMesh), initializer="random_normal",trainable= True , dtype=tf.float32)
 
    # x : shape (batch size, in_features) in grid
    # grid : the grid  [d, 2]  (dim, (min,max)
    def call(self, x ,grid ):
        # Position in [0,1]
        aPt = tf.cumsum(tf.nn.sigmoid(self.weiGhtaPt[:,:-1]), axis=-1)/tf.reduce_sum(tf.nn.sigmoid(self.weiGhtaPt),axis=-1,keepdims=True)
        # size (in_features, self.nbMesh-1)
        gridLocInt=  tf.tile(tf.expand_dims(grid[:,0], axis=-1), [1,self.nbMesh-1])  +  \
                                    tf.tile(tf.expand_dims( grid[:,1]-grid[:,0], axis=-1), [1,self.nbMesh-1])* aPt
        # size  (1,in_features, self.nbMesh+1)
        gridLocInt = tf.expand_dims(tf.concat([tf.expand_dims(grid[:,0], axis=-1),gridLocInt, tf.expand_dims(grid[:,1], axis=-1)] , axis=-1),axis=0)
        v0=  tf.nn.relu(1 -(x -gridLocInt[0,:,0])/(gridLocInt[0,:,1]-gridLocInt[0,:,0]))
        vInt = [ tf.where( x <=  gridLocInt[:,:,k], tf.nn.relu((x -gridLocInt[:,:,k-1])/(gridLocInt[:,:,k]-gridLocInt[:,:,k-1])), \
                         tf.nn.relu(1 -(x -gridLocInt[:,:,k])/(gridLocInt[:,:,k+1]-gridLocInt[:,:,k]))) for k in range(1,self.nbMesh)]
        v1=  tf.nn.relu((x -gridLocInt[:,:,-2])/(gridLocInt[:,:,-1]-gridLocInt[:,:,-2]))
        edgeVal = tf.stack([v0]+vInt+[v1], axis=-1) # (batchSize,in_features,self.nbMesh+1)
        values = tf.reduce_sum(tf.einsum("ijk,ljk->lij", self.weightsNet,edgeVal), axis=-1)                   
        gridOut= tf.stack([  tf.reduce_sum( tf.reduce_min(self.weightsNet, axis=-1), axis=-1), tf.reduce_sum( tf.reduce_max(self.weightsNet, axis=-1), axis=-1)], axis=-1)
        return values, gridOut
        
# Stack Kolmogorov layer to get a network
# Grid rang given as input for the network
class P1AdaptKan(tf.keras.Model):
    # inputDim    input   dimension
    # output_dim  output  dimension
    # layers      hiden layers
    # nbMesh      number of mesh per direction
    # grid        the grid for approximation
    # activation  activation function
    def __init__(self,  inputDim , output_dim, layers,
                 nbMesh,
                 grid ,
                 activation = tf.identity):
        
        super().__init__()
        self.inputDim=inputDim
        self.output_dim=  output_dim
        self.theLayers = [inputDim]+ layers + [output_dim]
        self.grid= tf.convert_to_tensor(grid, dtype =tf.float32)
        nbLay= len(self.theLayers)       
        self.layersOfNet =   [ P1AdaptLayer(self.theLayers[i], self.theLayers[i+1], nbMesh) for i in range(nbLay-1)]
        self.activation= activation
 
    def call(self, x):
        gridLoc =  self.grid
        for lay in self.layersOfNet:
            x, gridLoc= lay(x,gridLoc)
        return self.activation(x)
                   
class funcHard2D:
    def __init__(self):
        self.name_ = "funcHard2D"
        
    def __call__(self,x):
        x= 2*x-1
        phiI = 1 + tf.abs(x[:,0]) + x[:,1] + tf.sin(tf.abs(x[:,0])) + tf.tanh(x[:,1])
        phiAbove = phiI + (x[:,0] -x[:,1])
        phiBelow = phiI -  (x[:,0] -x[:,1]) + tf.sin(5*(x[:,0] -x[:,1]))
        return tf.where (x[:,1] > x[:,0], phiAbove, phiBelow)
    
class funcHardCos:
    def __init__(self,d):
        self.d =d
        self.name = "funcHardCos"
        
    def __call__(self,x):
        x= 2*x-1
        xInit = 0.5
        xLoc = xInit*tf.ones(shape=tf.shape(x))+ x/np.sqrt(self.d)
        nbSimul = tf.shape(x)[0]
        a =tf.convert_to_tensor(np.arange(1.,self.d+1.),dtype=tf.float32)
        xSum= tf.einsum("ij,j->i",xLoc,a)
        return tf.cos(xSum)
    
class func1D:
    def __init__(self):
       self.name_ = "func1D"
        
    def __call__(self,x):
        x= 2*x-1
        return  tf.sin(tf.abs(x[:,0]))
        
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--net', type=int, default=0)
    os.environ['CUDA_VISIBLE_DEVICES']='-1'
    args = parser.parse_args()
    net = args.net
    num_epochExt= 3000
    num_epoch= 100
    inputDim =4
    outputDim =1
    targetFunc = funcHardCos(inputDim)#funcHard2D() #func1D() #funcHard2D()

    batchSize = 10000

    @tf.function
    def minFunc( network, batchSize, targetFunc,inputDim ):
        x= tf.random.uniform([batchSize,inputDim ], minval=0, maxval =1)
        return tf.reduce_mean(tf.square(targetFunc(x)-network(x)[:,0]))

    print("Args" , args)
    if  net ==0:
        print("Enter P1Kan")
        nb_neurons = [20,20]
        nbMesh =10
        grid =np.stack([np.zeros([inputDim]), np.ones([inputDim])], axis=-1)
        network = P1AdaptKan(inputDim,outputDim,nb_neurons,nbMesh,grid )
    else:
        print("ENTER FFNet")
        nb_neurons = [40,40,40]
        network = tf.keras.Sequential([tf.keras.Input(shape=(inputDim,))] + \
                                      [layers.Dense( nb_neurons[i],activation= tf.nn.relu)  for i in range(len(nb_neurons)) ]+ \
                                      [layers.Dense(1)])

    lRate = 1e-3
    optimizer = tf.keras.optimizers.Adam(learning_rate = lRate)

    for  iout in range( num_epochExt):
        a=time.time()
        for epoch in range(num_epoch):
            with tf.GradientTape() as tape:               
                toOpt = minFunc(network, batchSize, targetFunc,inputDim)         
                gradients= tape.gradient(toOpt, network.trainable_variables)
                optimizer.apply_gradients(zip(gradients,network.trainable_variables))
        toOpt =  minFunc(network, batchSize*10, targetFunc,inputDim)
        print("Error", toOpt, "TI ", time.time()-a)
